import logging
import typing as t
from collections import Counter


class CounterHook:
    """A Seagrass hook that counts the number of times an event occurs.

    **Examples:**

    .. testsetup:: counterhook-examples

       from seagrass import Auditor
       from seagrass._docs import configure_logging

       configure_logging()
       auditor = Auditor()

    .. doctest:: counterhook-examples

       >>> from seagrass.hooks import CounterHook

       >>> hook = CounterHook()

       >>> event_a = auditor.create_event("event_a", hooks=[hook])

       >>> event_b = auditor.create_event("event_b", hooks=[hook])

       >>> with auditor.start_auditing(log_results=True):
       ...     for _ in range(15):
       ...         auditor.raise_event("event_a")
       ...     for _ in range(8):
       ...         auditor.raise_event("event_b")
       (INFO) seagrass: Calls to events recorded by CounterHook:
       (INFO) seagrass:     event_a: 15
       (INFO) seagrass:     event_b: 8
    """

    event_counter: t.Counter[str]

    def __init__(self) -> None:
        self.event_counter = Counter()

    def prehook(
        self, event_name: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> None:
        self.event_counter[event_name] += 1

    def posthook(self, event_name: str, result: t.Any, context: None) -> None:
        # Posthook does nothing
        pass

    def reset(self) -> None:
        self.event_counter.clear()

    def log_results(self, logger: logging.Logger) -> None:
        logger.info("Calls to events recorded by %s:", self.__class__.__name__)

        if len(self.event_counter) == 0:
            logger.info("    (no events recorded)")
            return

        for event in sorted(self.event_counter):
            logger.info("    %s: %d", event, self.event_counter[event])



# Basic errors that can be thrown while using Seagrass

import os
import typing as t


class SeagrassError(Exception):
    """A generic error for the Seagrass library."""


class EventNotFoundError(SeagrassError):
    """Raised when we try to reference an auditing event that does not currently
    exist."""

    event_name: str

    def __init__(self, event_name: str) -> None:
        """Create a new EventNotFoundError.

        :param str event_name: the name of the event that could not be found.
        """
        self.event_name = event_name
        msg = f"Audit event not found: {event_name}"
        super().__init__(msg)


class PosthookError(SeagrassError):
    """Raised when one or more Seagrass posthooks raised an error."""

    errors: t.List[Exception]

    def __init__(self, errors: t.List[Exception]) -> None:
        """Create a new PosthookError.

        :param List[Exception] errors: a list of errors that were raised.
        """
        self.errors = errors
        msg = f"{len(errors)} errors raised while processing posthooks and clean-up stages."
        msg += os.linesep
        msg += os.linesep.join(f"{e.__class__.__name__}: {e}" for e in errors)
        super().__init__(msg)



import sys
import typing as t
from enum import Enum, auto
from seagrass.base import ProtoHook, CleanupHook, prehook_priority, posthook_priority
from seagrass.errors import PosthookError

# A type variable used to represent the function wrapped by an Event.
F = t.Callable[..., t.Any]


class _HookContext(Enum):
    # Enum class used to represent cases where we don't have the context for a posthook
    MISSING = auto()

    def __repr__(self):
        return f"<{self.__class__.__name__}.{self.name}>"


class Event:
    """Defines an event that is under audit. The event wraps around a function; instead of calling
    the function, we call the event, which first triggers any prehooks, *then* calls the function,
    and then triggers posthooks."""

    # Use __slots__ since feasibly users may want to create a large
    # number of events
    __slots__ = [
        "func",
        "enabled",
        "name",
        "raise_runtime_events",
        "hooks",
        "prehook_audit_event_name",
        "posthook_audit_event_name",
        "__prehook_execution_order",
        "__posthook_execution_order",
    ]

    enabled: bool
    name: str
    raise_runtime_events: bool
    hooks: t.List[ProtoHook]
    prehook_audit_event_name: str
    posthook_audit_event_name: str
    __prehook_execution_order: t.List[int]
    __posthook_execution_order: t.List[int]

    def __init__(
        self,
        func: F,
        name: str,
        enabled: bool = True,
        hooks: t.List[ProtoHook] = [],
        raise_runtime_events: bool = False,
        prehook_audit_event_name: t.Optional[str] = None,
        posthook_audit_event_name: t.Optional[str] = None,
    ) -> None:
        """Create a new Event.

        :param Callable[[...],Any] func: the function being wrapped by this event.
        :param str name: the name of the event.
        :param bool enabled: whether to enable the event.
        :param List[ProtoHook] hooks: a list of all of the hooks that should be called whenever
            the event is triggered.
        :param bool raise_runtime_events: if ``True``, two `Python runtime audit events`_ are raised
            using `sys.audit`_ before and after running the function wrapped by the event.
        :param Optional[str] prehook_audit_event_name: the name of the runtime audit event
            that should be raised *before* calling the wrapped function. If set to ``None``,
            the audit event is automatically named ``f"prehook:{name}"``. This parameter is
            ignored if ``raise_runtime_events`` is ``False``.
        :param Optional[str] posthook_audit_event_name: the name of the runtime audit event
            that should be raised *after* calling the wrapped function. If set to ``None``,
            the audit event is automatically named ``f"posthook:{name}"``. This parameter is
            ignored if ``raise_runtime_events`` is ``False``.

        .. _Python runtime audit events: https://www.python.org/dev/peps/pep-0578/
        .. _sys.audit: https://docs.python.org/3/library/sys.html#sys.audit
        """
        self.func: F = func
        self.enabled = enabled
        self.name = name
        self.raise_runtime_events = raise_runtime_events
        self.hooks = []

        if prehook_audit_event_name is None:
            prehook_audit_event_name = f"prehook:{name}"
        if posthook_audit_event_name is None:
            posthook_audit_event_name = f"posthook:{name}"

        self.prehook_audit_event_name = prehook_audit_event_name
        self.posthook_audit_event_name = posthook_audit_event_name

        self.add_hooks(*hooks)

        # Set the order of execution for prehooks and posthooks.

    def add_hooks(self, *hooks: ProtoHook) -> None:
        """Add new hooks to the event.

        :param ProtoHook hooks: the hooks to add to the event.
        """
        for hook in hooks:
            self.hooks.append(hook)

        # Since we've updated the list of hooks, we need to re-determine the order
        # in which the hooks should be executed.
        self._set_hook_execution_order()

    def _set_hook_execution_order(self) -> None:
        """Determine the order in which the events' hooks should be executed."""
        # - Prehooks are ordered by ascending priority, then ascending list position
        # - Posthooks are ordered by descending priority, then descending list position
        self.__prehook_execution_order = sorted(
            range(len(self.hooks)), key=lambda i: (prehook_priority(self.hooks[i]), i)
        )
        self.__posthook_execution_order = sorted(
            range(len(self.hooks)),
            key=lambda i: (-posthook_priority(self.hooks[i]), -i),
        )

    def __call__(self, *args, **kwargs) -> t.Any:
        """Call the function wrapped by the Event. If the event is enabled, its prehooks and
        posthooks are executed before and after the execution of the wrapped function.

        :param args: the arguments to pass to the wrapped function.
        :param kwargs: the keyword arguments to pass to the wrapped function.
        """
        if not self.enabled:
            # We just return the result of the wrapped function
            return self.func(*args, **kwargs)

        if self.raise_runtime_events:
            sys.audit(self.prehook_audit_event_name, args, kwargs)

        # We use the exception_raised flag to tell us whether or not an exception was raised
        # in the course of executing the event and the hooks. If an exception *was* raised,
        # then we only call hooks' cleanup stages, and ignore the posthooks.
        exception_raised = False

        try:
            prehook_contexts = {}
            for hook_num in self.__prehook_execution_order:
                hook = self.hooks[hook_num]
                context = hook.prehook(self.name, args, kwargs)
                prehook_contexts[hook_num] = context

            result = self.func(*args, **kwargs)

        except Exception as ex:
            exception_raised = True
            raise ex

        finally:
            posthook_exceptions = []

            # Execute posthooks and cleanup stages by order of their priority.
            for hook_num in self.__posthook_execution_order:
                # In some cases (e.g., if a prehook raises an Exception), a context may
                # not exist for a given hook. We only execute the posthook and cleanup
                # if a context exists.
                context = prehook_contexts.get(hook_num, _HookContext.MISSING)
                if context != _HookContext.MISSING:
                    hook = self.hooks[hook_num]
                    if not exception_raised:
                        try:
                            hook.posthook(self.name, result, context)
                        except Exception as ex:
                            posthook_exceptions.append(ex)
                    if isinstance(hook, CleanupHook):
                        try:
                            hook.cleanup(self.name, context)
                        except Exception as ex:
                            posthook_exceptions.append(ex)

            # If one or more exceptions were thrown while processing the posthooks, we now
            # bubble up those errors.
            if len(posthook_exceptions) > 0:
                raise PosthookError(posthook_exceptions)

        if self.raise_runtime_events:
            sys.audit(self.posthook_audit_event_name, result)

        return result



# Testing utilities and base classes for testing Seagrass

import logging
import logging.config
import sys
import typing as t
from functools import wraps
from io import StringIO
from seagrass import Auditor
from seagrass.base import ProtoHook


class SeagrassTestCaseMixin:

    logging_output: StringIO
    logger: logging.Logger
    auditor: Auditor

    # Whether to emit logs that are >= WARNING to stdout
    log_warnings_to_stdout: bool = True

    def setUp(self) -> None:
        # Set up logging configuration
        self.logging_output = StringIO()

        handlers = ["default"]
        if self.log_warnings_to_stdout:
            handlers.append("warnings")

        config = {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "standard": {
                    "format": "(%(levelname)s) %(message)s",
                },
            },
            "handlers": {
                "default": {
                    "level": "DEBUG",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": self.logging_output,
                },
                # Add an additional handler for warnings that goes out to stdout. This way, if
                # any warnings/errors show up during tests, we can view them in the test output.
                "warnings": {
                    "level": "WARNING",
                    "formatter": "standard",
                    "class": "logging.StreamHandler",
                    "stream": "ext://sys.stdout",
                },
            },
            "loggers": {
                "test.seagrass": {
                    "handlers": handlers,
                    "level": "DEBUG",
                    "propagate": False,
                },
            },
        }

        logging.config.dictConfig(config)
        self.logger = logging.getLogger("test.seagrass")

        # Create a new auditor instance with the logger we just
        # set up
        self.auditor = Auditor(logger=self.logger)


class HookTestCaseMixin(SeagrassTestCaseMixin):
    """A base testing class for auditor hooks."""

    hook: ProtoHook
    hook_gen: t.Optional[t.Callable[[], ProtoHook]] = None

    # A list of all of the interfaces that the hook is expected
    # to satisfy.
    check_interfaces: t.Optional[t.Tuple[t.Type, ...]] = None

    @property
    def hook_name(self) -> str:
        return self.hook.__class__.__name__

    def setUp(self):
        super().setUp()

        if self.hook_gen is not None:
            self.hook = self.hook_gen()

    def test_hook_satisfies_interfaces(self):
        CheckableProtoHook = t.runtime_checkable(ProtoHook)
        self.assertIsInstance(
            self.hook,
            CheckableProtoHook,
            f"{self.hook_name} does not satisfy the ProtoHook interface",
        )

        interfaces = [] if self.check_interfaces is None else self.check_interfaces

        for interface in interfaces:
            self.assertIsInstance(
                self.hook,
                interface,
                f"{self.hook_name} does not satisfy the {interface.__name__} interface",
            )


F = t.Callable[..., None]


def req_python_version(
    min: t.Optional[t.Tuple[int, int]] = None, max: t.Optional[t.Tuple[int, int]] = None
) -> t.Callable[[F], F]:
    """Create a wrapper around a unit test to enforce a minimum
    or maximum Python version."""

    assert min is not None or max is not None
    version = sys.version_info

    err_msg = f"Test skipped: disabled for Python version = {version.major}.{version.minor}.{version.micro}, required"
    bad_version = False
    if (min is not None and max is not None) and (version < min or max < version):
        err_msg = f"{err_msg} {min[0]}.{min[1]} <= version <= {max[0]}.{max[1]}"
        bad_version = True
    elif (min is not None and max is None) and (version < min):
        err_msg = f"{err_msg} {min[0]}.{min[1]} <= version"
        bad_version = True
    elif (min is None and max is not None) and (version > max):
        err_msg = f"{err_msg} version <= {max[0]}.{max[1]}"
        bad_version = True

    if not bad_version:

        def decorator(test_func: F) -> F:
            return test_func

        return decorator
    else:

        def decorator(test_func: F) -> F:
            @wraps(test_func)
            def wrapper(self, *args, **kwargs):
                self.skipTest(err_msg)

            return wrapper

        return decorator



import sys
import typing as t
from abc import ABCMeta, abstractmethod
from contextvars import ContextVar, Token
from contextlib import contextmanager
from functools import wraps
from seagrass import get_audit_logger

# Type variable used to represent value returned from a function
R = t.TypeVar("R")


class RuntimeAuditHook(metaclass=ABCMeta):
    """Abstract base class that serves as a template for hooks whose body should be run as Python
    runtime audit hooks, in accordance with `PEP 578`_.

    **Examples:** in the code below, ``RuntimeEventCounterHook`` is a class derived from the
    ``RuntimeAuditHook`` base class that prints every time a runtime audit event is triggered.

    .. testsetup:: runtime-audit-hook-example

        from seagrass.auditor import Auditor
        auditor = Auditor()

    .. doctest:: runtime-audit-hook-example

        >>> import sys

        >>> from seagrass.hooks import RuntimeAuditHook

        >>> class RuntimeEventCounterHook(RuntimeAuditHook):
        ...     def __init__(self):
        ...         super().__init__()
        ...
        ...     def sys_hook(self, event, args):
        ...         print(f"Encountered {event=!r} with {args=}")
        ...

        >>> hook = RuntimeEventCounterHook()

        >>> @auditor.audit("my_event", hooks=[hook])
        ... def my_event(*args):
        ...     sys.audit("sys.my_event", *args)

        >>> with auditor.start_auditing():
        ...     my_event(42, "hello, world")
        Encountered event='sys.my_event' with args=(42, 'hello, world')

    .. _PEP 578: https://www.python.org/dev/peps/pep-0578/
    """

    # Default value of the propagate_errors attribute
    PROPAGATE_ERRORS_DEFAULT: t.Final[bool] = True

    # Whether to propagate errors from sys_hook.
    propagate_errors: bool = PROPAGATE_ERRORS_DEFAULT

    # A ContextVar that stores the latest event that's being executed
    _current_event_ctx: ContextVar[t.Optional[str]]

    # We keep _is_active and _current_event as hidden members of the class so that they appear as
    # read-only properties to child classes of RuntimeAuditHook.
    _is_active: bool
    _current_event: t.Optional[str]

    @abstractmethod
    def sys_hook(self, event: str, args: t.Tuple[t.Any, ...]) -> None:
        """The runtime auditing hook that gets executed every time sys.audit() is called.
        This must be defined in child classes of RuntimeAuditHook."""

    @property
    def is_active(self) -> bool:
        """Return whether or not the hook is currently active (i.e., whether a Seagrass event that
        uses the hook is currently executing.)"""
        return self._is_active

    @property
    def current_event(self) -> t.Optional[str]:
        """Returns the current Seagrass event being executed that's hooked by this function. If no
        events using this hook are being executed, ``current_event`` is ``None``."""
        return self._current_event

    def _update_properties(self) -> None:
        """Update the ``current_event`` and ``is_active`` properties."""
        self._current_event = self._current_event_ctx.get()
        self._is_active = self.current_event is not None

    def _update_decorator(func: t.Callable[..., R]) -> t.Callable[..., R]:  # type: ignore[misc]
        # NOTE: mypy will flag this as erroneous because it is a non-static method that doesn't
        # include the argument 'self'
        # Ref: https://github.com/python/mypy/issues/7778
        """Function decorator that causes functions to reset the current_event and is_active
        properties every time it gets called."""

        @wraps(func)
        def wrapper(self, *args, **kwargs):
            try:
                return func(self, *args, **kwargs)
            finally:
                self._update_properties()

        return wrapper

    def __init__(self, propagate_errors: t.Optional[bool] = None) -> None:
        self._current_event_ctx = ContextVar("current_event", default=None)
        if propagate_errors is not None:
            self.propagate_errors = propagate_errors
        self._update_properties()

        # Add the runtime audit hook after initializing the properties since the hook will in most
        # cases use some of the properties of the RuntimeAuditHook.
        sys.addaudithook(self._sys_hook)

    def _sys_hook(self, event: str, args: t.Tuple[t.Any, ...]) -> None:
        """A wrapper around the sys_hook abstract method that first checks whether the hook
        is currently active before it executes anything. This is the function that actually
        gets added with sys.addaudithook, not sys_hook."""
        if self.is_active:
            try:
                self.sys_hook(event, args)
            except Exception as ex:
                if self.propagate_errors:
                    raise ex
                else:
                    if (logger := get_audit_logger()) is not None:
                        # Temporarily disable the hook, since emitting a log could create new
                        # runtime events. In some cases this could lead to an infinite recursion.
                        with self._disable_runtime_hook():
                            logger.error(
                                "%s raised in %s.sys_hook: %s",
                                ex.__class__.__name__,
                                self.__class__.__name__,
                                ex,
                            )

    @contextmanager
    def _disable_runtime_hook(self) -> t.Iterator[None]:
        """Temporarily the runtime hook."""
        is_active = self._is_active
        self._is_active = False
        try:
            yield None
        finally:
            self._is_active = is_active

    @_update_decorator
    def prehook(
        self, event: str, args: t.Tuple[t.Any, ...], kwargs: t.Dict[str, t.Any]
    ) -> Token:
        return self._current_event_ctx.set(event)

    def posthook(self, event: str, result: t.Any, context: Token) -> None:
        pass

    @_update_decorator
    def cleanup(self, event: str, context: Token) -> None:
        self._current_event_ctx.reset(context)


RuntimeAuditHook.__init__.__doc__ = f"""\
Initialize the RuntimeAuditHook.

:param Optional[bool] propagate_errors: if not equal to ``None``, set the ``propagate_errors``
    attribute of the hook, which controls whether errors raised by :py:meth:`sys_hook` are
    propagated. Hooks for which ``propagate_errors`` is ``False`` will log errors raised by
    :py:meth:`sys_hook`, but won't raise them.

    By default, ``hook.propagate_errors = {RuntimeAuditHook.PROPAGATE_ERRORS_DEFAULT}`` for
    RuntimeAuditHooks.
"""



# Tests for Auditor creation and basic functionality

import logging
import unittest
from io import StringIO
from seagrass import Auditor
from seagrass.errors import EventNotFoundError
from test.utils import SeagrassTestCaseMixin


class CreateAuditorTestCase(unittest.TestCase):
    """Tests for creating a new Auditor instance."""

    def _clear_logging_output(self):
        # Helper function to clear output buffer used for testing logging
        self.logging_output.seek(0)
        self.logging_output.truncate()

    def _configure_logger(self, logger: logging.Logger):
        # Set the testing configuration for an input logger
        logger.setLevel(logging.INFO)

        fh = logging.StreamHandler(self.logging_output)
        fh.setLevel(logging.DEBUG)

        formatter = logging.Formatter("(%(levelname)s) %(name)s: %(message)s")
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    def setUp(self):
        self.logging_output = StringIO()

    def test_default_logger_used_by_auditor(self):
        # By default, Auditor should use the "seagrass" logger if no logger is specified
        self._configure_logger(logging.getLogger("seagrass"))

        auditor = Auditor()
        auditor.logger.info("Hello, world!")
        auditor.logger.debug("This message shouldn't appear")

        output = self.logging_output.getvalue()
        self.assertEqual(output, "(INFO) seagrass: Hello, world!\n")

    def test_use_custom_logger_for_auditor(self):
        # If a logger is explicitly specified, Auditor should use that logger instead
        # of the default. The logger can either be provided as a logging.Logger
        # instance or as the name of a logger.
        self._configure_logger(logging.getLogger("test_logger"))

        for logger in ("test_logger", logging.getLogger("test_logger")):
            auditor = Auditor(logger=logger)
            auditor.logger.info("Hello, world!")
            auditor.logger.debug("This message shouldn't appear")

            output = self.logging_output.getvalue()
            self.assertEqual(output, "(INFO) test_logger: Hello, world!\n")

            self._clear_logging_output()


class SimpleAuditorFunctionsTestCase(SeagrassTestCaseMixin, unittest.TestCase):
    def test_define_new_event(self):
        # Define a new event and ensure that it gets added to the auditor's events
        # dictionary and event_wrapper dictionary
        @self.auditor.audit("test.foo")
        def foo():
            return

        self.assertIn("test.foo", self.auditor.events)
        self.assertIn("test.foo", self.auditor.event_wrappers)

    def test_define_two_events_with_the_same_name(self):
        @self.auditor.audit("test.foo")
        def foo_1():
            return

        with self.assertRaises(ValueError):

            @self.auditor.audit("test.foo")
            def foo_2():
                return

    def test_create_empty_event(self):
        # Create a new audit event that doesn't wrap any existing function.
        with self.assertRaises(EventNotFoundError):
            self.auditor.raise_event("test.signal", 1, 2, name="Alice")

        class TestHook:
            def __init__(self):
                self.last_prehook_args = self.last_posthook_args = None

            def prehook(self, event_name, args, kwargs):
                self.last_prehook_args = (event_name, args, kwargs)

            def posthook(self, event_name, result, context):
                self.last_posthook_args = (event_name, result)

        hook = TestHook()
        self.auditor.create_event("test.signal", hooks=[hook])

        # Event shouldn't be triggered outside of an auditing context
        self.auditor.raise_event("test.signal", 1, 2, name="Alice")
        self.assertEqual(hook.last_prehook_args, None)
        self.assertEqual(hook.last_posthook_args, None)

        with self.auditor.start_auditing():
            self.auditor.raise_event("test.signal", 1, 2, name="Alice")

        self.assertEqual(
            hook.last_prehook_args, ("test.signal", (1, 2), {"name": "Alice"})
        )
        self.assertEqual(hook.last_posthook_args, ("test.signal", None))

    def test_raise_event_cumsum(self):
        # Insert an audit event into the function my_sum so that we can monitor the internal
        # state of the function as it's executing. In this case, we'll be retrieving the
        # cumulative sum at various points in time.
        def my_sum(*args):
            total = 0.0
            for arg in args:
                self.auditor.raise_event("my_sum.cumsum", total)
                total += arg

        class MySumHook:
            def __init__(self):
                self.cumsums = []

            def prehook(self, event_name, args, kwargs):
                self.cumsums.append(args[0])

            def posthook(self, *args):
                pass

        hook = MySumHook()
        self.auditor.create_event("my_sum.cumsum", hooks=[hook])

        with self.auditor.start_auditing():
            my_sum(1, 2, 3, 4)

        self.assertEqual(hook.cumsums, [0.0, 1.0, 3.0, 6.0])


if __name__ == "__main__":
    unittest.main()



# Tests for the LoggerHook auditing hook.

import logging
import unittest
from functools import reduce
from operator import add, mul
from seagrass.hooks import LoggingHook
from test.utils import HookTestCaseMixin


class LoggingHookTestCase(HookTestCaseMixin, unittest.TestCase):
    def setUp(self):
        super(HookTestCaseMixin, self).setUp()

        self.hook_pre = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"hook_pre: {e}, {args=}, {kwargs=}",
        )
        self.hook_both = LoggingHook(
            prehook_msg=lambda e, args, kwargs: f"hook_both: {e}, {args=}, {kwargs=}",
            posthook_msg=lambda e, result: f"hook_both: {e}, {result=}",
            loglevel=logging.INFO,
        )

        # Use hook_both as self.hook for running additional tests that are defined for
        # test case classes that subclass from HookTestMixin.
        self.hook = self.hook_both

    def test_hook_function(self):
        event = "test.multiply_or_add"

        @self.auditor.audit(event, hooks=[self.hook_pre, self.hook_both])
        def multiply_or_add(*args, op="*"):
            if op == "*":
                return reduce(mul, args, 1)
            elif op == "+":
                return reduce(add, args, 0)
            else:
                raise ValueError(f"Unknown operation '{op}'")

        args = (1, 2, 3, 4)
        kwargs_add = {"op": "+"}
        with self.auditor.start_auditing():
            multiply_or_add(*args)
            multiply_or_add(*args, **kwargs_add)

        output = self.logging_output.getvalue().rstrip().split("\n")
        self.assertEqual(output[0], f"(DEBUG) hook_pre: {event}, {args=}, kwargs={{}}")
        self.assertEqual(output[1], f"(INFO) hook_both: {event}, {args=}, kwargs={{}}")
        self.assertEqual(output[2], f"(INFO) hook_both: {event}, result={24}")
        self.assertEqual(
            output[3], f"(DEBUG) hook_pre: {event}, {args=}, kwargs={kwargs_add}"
        )
        self.assertEqual(
            output[4], f"(INFO) hook_both: {event}, {args=}, kwargs={kwargs_add}"
        )
        self.assertEqual(output[5], f"(INFO) hook_both: {event}, result={10}")



# Tests for the CounterHook auditing hook.

from seagrass.base import ResettableHook, LogResultsHook
from seagrass.hooks import CounterHook
from test.utils import HookTestCaseMixin
import unittest


class CounterHookTestCase(HookTestCaseMixin, unittest.TestCase):

    hook_gen = CounterHook
    check_interfaces = (ResettableHook, LogResultsHook)

    def test_hook_function(self):
        @self.auditor.audit("test.say_hello", hooks=[self.hook])
        def say_hello(name: str) -> str:
            return f"Hello, {name}!"

        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

        # Hook should not get called outside of an auditing context
        say_hello("Alice")
        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

        with self.auditor.start_auditing():
            for name in ("Alice", "Bob", "Cathy"):
                say_hello(name)
        self.assertEqual(self.hook.event_counter["test.say_hello"], 3)
        self.assertEqual(set(self.hook.event_counter), set(("test.say_hello",)))

        # Upon resetting the hook, all event counts should be set back to zero
        self.hook.reset()
        self.assertEqual(self.hook.event_counter["test.say_hello"], 0)

    def test_counter_hook_logging(self):
        # Collect event information with CounterHook and then check the logs
        # that are emitted by CounterHook.log_results.

        self.auditor.create_event("event_b", hooks=[self.hook])
        self.auditor.create_event("event_a", hooks=[self.hook])
        self.auditor.create_event("event_c", hooks=[self.hook])

        with self.auditor.start_auditing():
            for _ in range(904):
                self.auditor.raise_event("event_b")
            for _ in range(441):
                self.auditor.raise_event("event_a")
            for _ in range(58):
                self.auditor.raise_event("event_c")

        self.auditor.log_results()
        lines = self.logging_output.getvalue().rstrip().split("\n")
        self.assertEqual(len(lines), 4)
        self.assertEqual(lines[0], "(INFO) Calls to events recorded by CounterHook:")
        self.assertEqual(lines[1], "(INFO)     event_a: 441")
        self.assertEqual(lines[2], "(INFO)     event_b: 904")
        self.assertEqual(lines[3], "(INFO)     event_c: 58")


if __name__ == "__main__":
    unittest.main()



# This module primarily exists for the purposes of helping generate the
# Seagrass documentation. The code in this module is not part of the
# public Seagrass API.

import logging
import logging.config
from seagrass import DEFAULT_LOGGER_NAME


def configure_logging(name: str = DEFAULT_LOGGER_NAME) -> logging.Logger:
    """Set up the default logging configuration for the documentation."""

    logger = logging.getLogger(name)
    while logger.hasHandlers():
        logger.removeHandler(logger.handlers[0])

    logging.config.dictConfig({
        "version": 1,
        "disable_existing_loggers": False,
        "formatters": {
            "standard": {
                "format": "(%(levelname)s) %(name)s: %(message)s",
            },
        },
        "handlers": {
            "default": {
                "level": "DEBUG",
                "formatter": "standard",
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
            },
        },
        "loggers": {
            DEFAULT_LOGGER_NAME: {
                "handlers": ["default"],
                "level": "DEBUG",
                "propagate": False,
            },
        }
    })

    return logging.getLogger(DEFAULT_LOGGER_NAME)



# Tests for the FileOpenHook auditing hook.

import tempfile
import unittest
from test.utils import HookTestCaseMixin
from seagrass.base import LogResultsHook, ResettableHook, CleanupHook
from seagrass.hooks import FileOpenHook


class FileOpenHookTestCase(HookTestCaseMixin, unittest.TestCase):

    check_interfaces = (LogResultsHook, ResettableHook, CleanupHook)

    @staticmethod
    def hook_gen():
        return FileOpenHook()

    def test_hook_function(self):
        @self.auditor.audit("test.say_hello", hooks=[self.hook])
        def say_hello(filename, name) -> str:
            with open(filename, "w") as f:
                f.write(f"Hello, {name}!\n")

            with open(filename, "r") as f:
                return f.read()

        with tempfile.NamedTemporaryFile() as f:
            # Even though we're using sys.audit hooks, calls to say_hello should not
            # trigger the audit hook unless we're in an auditing context.
            say_hello(f.name, "Alice")
            with self.auditor.start_auditing():
                result = say_hello(f.name, "Alice")
            say_hello(f.name, "Alice")

            self.assertEqual(result, "Hello, Alice!\n")

            keys = list(self.hook.file_open_counter["test.say_hello"].keys())
            keys.sort(key=lambda info: info.mode)
            self.assertEqual(keys[0].filename, f.name)
            self.assertEqual(keys[0].mode, "r")
            self.assertEqual(keys[1].filename, f.name)
            self.assertEqual(keys[1].mode, "w")

            # Check the logging output
            self.auditor.log_results()
            self.logging_output.seek(0)
            lines = [line.rstrip() for line in self.logging_output.readlines()]

            # Header line + one line for one event + two lines for one read of the temporary
            # file, one write to it
            self.assertEqual(len(lines), 4)

    def test_nested_calls_to_hooked_functions(self):
        # Opens that occur in an event that's nested in another event should only count for
        # the inner event, not the outer one.
        @self.auditor.audit("test.readlines", hooks=[self.hook])
        def readlines(filename):
            with open(filename, "r") as f:
                return f.readlines()

        @self.auditor.audit("test.tabify", hooks=[self.hook])
        def tabify(filename, outfile):
            lines = readlines(filename)
            with open(outfile, "w") as f:
                for line in lines:
                    f.write("\t" + line)

        with tempfile.NamedTemporaryFile() as tf1:
            with tempfile.NamedTemporaryFile() as tf2:
                with self.auditor.start_auditing():
                    with open(tf1.name, "w") as f:
                        f.write("hello\nworld!")
                    tabify(tf1.name, tf2.name)

                self.assertEqual(
                    sorted(self.hook.file_open_counter.keys()),
                    ["test.readlines", "test.tabify"],
                )

                readlines_keys = list(
                    self.hook.file_open_counter["test.readlines"].keys()
                )
                tabify_keys = list(self.hook.file_open_counter["test.tabify"].keys())

                # Opened one file for reading in readlines
                # Opened one file for writing in tabify
                self.assertEqual(len(readlines_keys), 1)
                self.assertEqual(len(tabify_keys), 1)

                # Check statistics about the file that was read
                read_info = readlines_keys[0]
                self.assertEqual(read_info.filename, tf1.name)
                self.assertEqual(read_info.mode, "r")
                self.assertEqual(
                    self.hook.file_open_counter["test.readlines"][read_info], 1
                )
                self.assertEqual(
                    self.hook.file_open_counter["test.tabify"][read_info], 0
                )

                # Check statistics about the file that was written to
                write_info = tabify_keys[0]
                self.assertEqual(write_info.filename, tf2.name)
                self.assertEqual(write_info.mode, "w")
                self.assertEqual(
                    self.hook.file_open_counter["test.readlines"][write_info], 0
                )
                self.assertEqual(
                    self.hook.file_open_counter["test.tabify"][write_info], 1
                )


if __name__ == "__main__":
    unittest.main()
